import {DownloadSpeedMeter} from '../../core/DownloadSpeedMeter';
import {Event} from '../../enums/Event';
import {MIMETypes} from '../../enums/MIMETypes';
import {Player} from '../../enums/Player';
import {AnalyticsStateMachineOptions} from '../../types/AnalyticsStateMachineOptions';
import {DownloadSpeedInfo} from '../../types/DownloadSpeedInfo';
import {PlaybackInfo} from '../../types/PlaybackInfo';
import {QualityLevelInfo} from '../../types/QualityLevelInfo';
import {SubtitleInfo} from '../../types/SubtitleInfo';
import {HlsSpeedMeterAdapter} from './HlsSpeedMeterAdapter';
import {HTML5InternalAdapter} from './HTML5InternalAdapter';
import {InternalAdapterAPI} from './InternalAdapterAPI';

export class HlsInternalAdapter extends HTML5InternalAdapter implements InternalAdapterAPI {
  get downloadSpeedInfo(): DownloadSpeedInfo {
    return this.speedMeter.getInfo();
  }

  private speedMeter: DownloadSpeedMeter;

  constructor(private hls: Hls, opts?: AnalyticsStateMachineOptions) {
    super(null, opts);
    this.resetMedia();
    this.speedMeter = new HlsSpeedMeterAdapter(hls, new DownloadSpeedMeter()).getDownloadSpeedMeter();
  }

  public initialize() {
    super.initialize();
    this.registerHlsEvents();
  }

  public clearValues(): void {
    this.speedMeter.reset();
  }

  public getPlayerName = () => Player.HLSJS;
  public getPlayerVersion = () => (this.hls as any).constructor.version;

  public getCurrentQualityLevelInfo(): QualityLevelInfo | null {
    if (!this.hls.levels) {
      return null;
    }
    const currentLevelObj = this.hls.levels[this.hls.currentLevel];
    if (!currentLevelObj) {
      return null;
    }

    const bitrate = currentLevelObj.bitrate;
    const width = currentLevelObj.width;
    const height = currentLevelObj.height;

    return {
      bitrate,
      width,
      height,
    };
  }

  /**
   * @override
   */
  public isLive = () => {
    const hls = this.hls;
    if (!hls.levels || !hls.currentLevel) {
      return false;
    }
    if (hls.currentLevel < 0) {
      return false;
    }
    const currentLevelObj = hls.levels[hls.currentLevel];
    if (!currentLevelObj) {
      return false;
    }
    return currentLevelObj.details.live;
  };

  /**
   * @override
   */
  public getMIMEType() {
    return MIMETypes.HLS;
  }

  /**
   * @override
   */
  public getStreamURL() {
    // @ts-ignore
    return this.hls.url;
  }

  public registerHlsEvents() {
    const hls = this.hls as any;

    hls.on(hls.constructor.Events.MEDIA_ATTACHING, this.onMediaAttaching.bind(this));
    hls.on(hls.constructor.Events.MEDIA_DETACHING, this.onMediaDetaching.bind(this));
    hls.on(hls.constructor.Events.MANIFEST_LOADING, this.onManifestLoading.bind(this));
    hls.on(hls.constructor.Events.AUDIO_TRACK_SWITCHED, this.onAudioTrackSwitched.bind(this));
    hls.on(hls.constructor.Events.SUBTITLE_TRACK_SWITCH, this.onSubtitleLanguageSwitched.bind(this));
    hls.on(hls.constructor.Events.ERROR, this.onHlsError.bind(this));

    // media is already attached, event has been triggered before
    // or we are in the event handler of this event itself.
    // we can not know how the stacktrace to this constructor will look like.
    // therefore we will guard from this case in
    // the onMediaAttaching method (avoid running it twice)
    if (hls.media) {
      this.onMediaAttaching();
      this.onManifestLoading();
    }
  }

  public onMediaAttaching() {
    // in case we are called again (when we are triggering this ourselves
    // but from the event handler of MEDIA_ATTACHING) we should not run again.
    if (this.mediaElement) {
      return;
    }

    this.mediaElement = this.hls.media;

    this.registerMediaElement();
    this.onMaybeReady();
  }

  public onMediaDetaching() {
    this.unregisterMediaElement();
  }

  public onManifestLoading() {
    this.onMaybeReady();
  }

  public getCurrentPlaybackInfo(): PlaybackInfo {
    const selectedSubtitle = this.getSelectedSubtitleLanguage();

    const info: PlaybackInfo = {
      ...super.getCurrentPlaybackInfo(),
      audioLanguage: this.getSelectedAudioLanguage(),
      subtitleEnabled: selectedSubtitle != null ? selectedSubtitle.enabled : undefined,
      subtitleLanguage: selectedSubtitle != null ? selectedSubtitle.language : undefined,
    };

    return info;
  }

  /**
   * errorData: { type : error type, details : error details, fatal : if true, hls.js cannot/will not try to recover, if false, hls.js will try to recover,other error specific data }
   * https://hls-js.netlify.com/api-docs/file/src/events.js.html#lineNumber98
   */
  public onHlsError(errorName, errorData) {
    const isFatal = errorData ? errorData.fatal : true;
    if (isFatal) {
      const mediaElement = this.mediaElement;
      let currentTime;
      if (mediaElement != null) {
        currentTime = mediaElement.currentTime;
      }

      const errorMessage = errorData ? `${errorData.type}: ${errorData.details}` : null;
      this.eventCallback(Event.ERROR, {
        currentTime,
        code: this.getErrorCodeFromHlsErrorType(errorData.type),
        message: errorMessage,
      });
    }
  }

  public onAudioTrackSwitched(event: string, data: {id: string}) {
    const mediaElement = this.mediaElement;
    let currentTime;
    if (mediaElement != null) {
      currentTime = mediaElement.currentTime;
    }

    this.eventCallback(Event.AUDIOTRACK_CHANGED, {
      currentTime,
    });
  }

  public onSubtitleLanguageSwitched(event: string, data: {id: string}) {
    const mediaElement = this.mediaElement;
    let currentTime;
    if (mediaElement != null) {
      currentTime = mediaElement.currentTime;
    }

    this.eventCallback(Event.SUBTITLE_CHANGE, {
      currentTime,
    });
  }

  /**
   * returns mapped error code for Hlsjs ErrorTypes
   * @param type one of the ErrorTypes according to https://github.com/video-dev/hls.js/blob/master/src/errors.ts
   */
  private getErrorCodeFromHlsErrorType(type: string): number {
    switch (type) {
      case 'networkError':
        return 2;
      case 'mediaError':
        return 3;
      case 'keySystemError':
        return 4;
      case 'muxError':
        return 5;
      // default equals `otherError`
      default:
        return 1;
    }
  }

  private getSelectedAudioLanguage(): string | undefined {
    if (this.hls.audioTrack == null || this.hls.audioTrack < 0) {
      return undefined;
    }
    return (this.hls.audioTracks[this.hls.audioTrack] as any).lang;
  }

  private getSelectedSubtitleLanguage(): SubtitleInfo | undefined {
    // Check if function is supported (in latest hls version), otherwise use fallback option
    if (this.hls.subtitleDisplay != null) {
      const isSubtitleDisplayed = this.hls.subtitleTrack >= 0 && this.hls.subtitleDisplay === true;
      return {
        enabled: isSubtitleDisplayed,
        language: isSubtitleDisplayed ? this.hls.subtitleTracks[this.hls.subtitleTrack].lang : undefined,
      };
    } else {
      const {subtitleTrackController} = this.hls as any;
      if (subtitleTrackController != null && subtitleTrackController.media != null) {
        return this.getSelectedSubtitleFromMediaElement(subtitleTrackController.media);
      }
    }
    return undefined;
  }
}
