import {LicenseCheckingBackend} from '../js/core/LicenseCheckingBackend';
import {NoOpBackend} from '../js/core/NoOpBackend';
import {QueueBackend} from '../js/core/QueueBackend';
import {RemoteBackend} from '../js/core/RemoteBackend';
import {LicensingRequest, LicensingResponse, LicensingResult} from '../js/types/LicensingRequest';
import {AnalyticsCall} from '../js/utils/AnalyticsCall';
import {logger} from '../js/utils/Logger';
import {ANALYTICS_BACKEND_BASE_URL} from '../js/utils/Settings';
import {noOp} from '../js/utils/Utils';
import {MockDeferredLicenseLoadingAdapter} from './utils/MockDeferredLicenseLoadingAdapter';
import {MockDeferredLicenseLoadingAdapterWithSupport} from './utils/MockDeferredLicenseLoadingAdapterWithSupport';

jest.mock('../js/utils/Logger');
jest.mock('../js/utils/AnalyticsCall');

const granted: LicensingResponse = {
  status: LicensingResult.Granted,
  message: 'here you go',
};

const denied = {
  status: LicensingResult.Denied,
  message: 'You shall not pass!',
};

const validLicensingRequest: LicensingRequest = {
  domain: 'foo.com',
  key: '81354CDA-0EE6-4EDB-9A12-9EC7F80BBFE8',
  version: `1.2.3`,
};

const licensingRequestNoKey: LicensingRequest = {
  domain: 'foo.com',
  version: `1.2.3`,
};

const mockGlobalProperty = (globalObject) => (key) => (value) => {
  // save original implementation in order to unmock later
  const original = globalObject[key];

  // mock key on the global object
  Object.defineProperty(globalObject, key, {value, writable: true});

  // return an unmock function restoring original implementation
  return () => mockGlobalProperty(globalObject)(key)(original);
};

beforeEach(() => {
  (logger.errorMessageToUser as any).mockReset();
});

describe('LicenseCheckingBackend', () => {
  describe('License key provided', () => {
    const adapter = new MockDeferredLicenseLoadingAdapter();
    test('should call licenseCall and not subscribe to adapter events', () => {
      const mockLicenseCall = jest.fn().mockResolvedValue(granted);
      const backend = new LicenseCheckingBackend(
        validLicensingRequest,
        mockLicenseCall,
        ANALYTICS_BACKEND_BASE_URL,
        adapter
      );
      expect(mockLicenseCall).toHaveBeenCalledTimes(1);
      expect(adapter.onLicenseCallFailed.subscriberCount).toBe(0);
      expect(adapter.onLicenseKeyReceived.subscriberCount).toBe(0);
    });
    describe('Successful response', () => {
      test('it queues requests until the LicenseCall resolves', () => {
        let trigger;
        const mockLicenseCall = (key: string, domain: string, version: string) => {
          return new Promise<LicensingResponse>((resolve) => {
            trigger = resolve;
          });
        };
        const backend = new LicenseCheckingBackend(
          validLicensingRequest,
          mockLicenseCall,
          ANALYTICS_BACKEND_BASE_URL,
          adapter
        );
        expect((backend as any).backend).toBeInstanceOf(QueueBackend);

        trigger(granted);

        const promise: any = backend.promise;
        expect(promise).toBeDefined();
        return promise.then((result) => {
          expect((backend as any).backend).toBeInstanceOf(RemoteBackend);
        });
      });
      test('it calls flush on queue once license call is granted', () => {
        let trigger;
        const mockLicenseCall = (key: string, domain: string, version: string) => {
          return new Promise<LicensingResponse>((resolve) => {
            trigger = resolve;
          });
        };
        const backend = new LicenseCheckingBackend(
          validLicensingRequest,
          mockLicenseCall,
          ANALYTICS_BACKEND_BASE_URL,
          adapter
        );

        const backendBackend = (backend as any).backend;
        expect(backendBackend).toBeInstanceOf(QueueBackend);
        const flushToMock = jest.fn();
        (backendBackend as QueueBackend).flushTo = flushToMock;
        trigger(granted);

        const promise: any = backend.promise;
        expect(promise).toBeDefined();
        return promise.then((result) => {
          expect(flushToMock).toHaveBeenCalledTimes(1);
          expect(flushToMock).toHaveBeenCalledWith(expect.any(RemoteBackend));
        });
      });
    });
    describe('rejected response', () => {
      test('it does not resolve', () => {
        const mockLicenseCall = jest.fn().mockRejectedValue(denied);
        const backend = new LicenseCheckingBackend(
          validLicensingRequest,
          mockLicenseCall,
          ANALYTICS_BACKEND_BASE_URL,
          adapter
        );

        return mockLicenseCall()
          .then(() => {
            expect(true).toEqual(false);
          })
          .catch((msg) => {
            expect(msg.message).toBeDefined();
            expect(msg.status).toEqual('denied');
          });
      });
      test('it assigns a noop backend', () => {
        const mockLicenseCall = jest.fn().mockRejectedValue(denied);
        const backend = new LicenseCheckingBackend(
          validLicensingRequest,
          mockLicenseCall,
          ANALYTICS_BACKEND_BASE_URL,
          adapter
        );

        const promise: any = backend.promise;
        expect(promise).toBeDefined();
        return promise.then(() => {
          expect((backend as any).backend).toBeInstanceOf(NoOpBackend);
        });
      });
      test('it logs to the console', () => {
        const mockLicenseCall = jest.fn().mockRejectedValue(denied);
        const backend = new LicenseCheckingBackend(
          validLicensingRequest,
          mockLicenseCall,
          ANALYTICS_BACKEND_BASE_URL,
          adapter
        );

        const promise: any = backend.promise;
        expect(promise).toBeDefined();
        return promise.then(() => {
          expect(logger.errorMessageToUser).toHaveBeenCalledTimes(1);
        });
      });
    });
  });
  describe('No license key provided', () => {
    describe('Deferred loading not supported', () => {
      test('should not call licenseCall and not subscribe to adapter events (support for deferred loading = undefined)', () => {
        const adapter = new MockDeferredLicenseLoadingAdapter();
        const mockLicenseCall = jest.fn().mockResolvedValue(granted);
        const backend = new LicenseCheckingBackend(
          licensingRequestNoKey,
          mockLicenseCall,
          ANALYTICS_BACKEND_BASE_URL,
          adapter
        );
        expect(mockLicenseCall).toHaveBeenCalledTimes(0);
        expect(adapter.onLicenseCallFailed.subscriberCount).toBe(0);
        expect(adapter.onLicenseKeyReceived.subscriberCount).toBe(0);
        expect((backend as any).backend).toBeInstanceOf(NoOpBackend);
      });

      test('should not call licenseCall and not subscribe to adapter events (support for deferred loading = false)', () => {
        const adapter = new MockDeferredLicenseLoadingAdapterWithSupport(false);
        const mockLicenseCall = jest.fn().mockResolvedValue(granted);
        const backend = new LicenseCheckingBackend(
          licensingRequestNoKey,
          mockLicenseCall,
          ANALYTICS_BACKEND_BASE_URL,
          adapter
        );
        expect(mockLicenseCall).toHaveBeenCalledTimes(0);
        expect(adapter.onLicenseCallFailed.subscriberCount).toBe(0);
        expect(adapter.onLicenseKeyReceived.subscriberCount).toBe(0);
        expect((backend as any).backend).toBeInstanceOf(NoOpBackend);
      });
    });

    describe('Deferred loading supported', () => {
      jest.useFakeTimers();

      test('should not call licenseCall and should subscribe to adapter events', () => {
        const adapter = new MockDeferredLicenseLoadingAdapterWithSupport(true);
        const mockLicenseCall = jest.fn().mockResolvedValue(granted);
        const backend = new LicenseCheckingBackend(
          licensingRequestNoKey,
          mockLicenseCall,
          ANALYTICS_BACKEND_BASE_URL,
          adapter
        );
        expect(mockLicenseCall).toHaveBeenCalledTimes(0);
        expect(adapter.onLicenseCallFailed.subscriberCount).toBe(1);
        expect(adapter.onLicenseKeyReceived.subscriberCount).toBe(1);
        expect((backend as any).backend).toBeInstanceOf(QueueBackend);
      });

      test('should not call licenseCall and should time out if no events from adapter received', async () => {
        const adapter = new MockDeferredLicenseLoadingAdapterWithSupport(true);
        const mockLicenseCall = jest.fn().mockResolvedValue(granted);
        const backend = new LicenseCheckingBackend(
          licensingRequestNoKey,
          mockLicenseCall,
          ANALYTICS_BACKEND_BASE_URL,
          adapter
        );

        expect(mockLicenseCall).toHaveBeenCalledTimes(0);
        expect(adapter.onLicenseCallFailed.subscriberCount).toBe(1);
        expect(adapter.onLicenseKeyReceived.subscriberCount).toBe(1);
        expect((backend as any).backend).toBeInstanceOf(QueueBackend);

        jest.runAllTimers();

        expect(mockLicenseCall).toHaveBeenCalledTimes(0);
        expect(adapter.onLicenseCallFailed.subscriberCount).toBe(0);
        expect(adapter.onLicenseKeyReceived.subscriberCount).toBe(0);
        expect((backend as any).backend).toBeInstanceOf(NoOpBackend);
      });

      describe('License key couldn`t retrieval failed at adapter', () => {
        test('should not call licenseCall, should unsubscribe from adapter and should clear timeout', () => {
          const adapter = new MockDeferredLicenseLoadingAdapterWithSupport(true);
          const mockLicenseCall = jest.fn().mockResolvedValue(granted);
          const backend = new LicenseCheckingBackend(
            licensingRequestNoKey,
            mockLicenseCall,
            ANALYTICS_BACKEND_BASE_URL,
            adapter
          );

          expect(mockLicenseCall).toHaveBeenCalledTimes(0);
          expect(adapter.onLicenseCallFailed.subscriberCount).toBe(1);
          expect(adapter.onLicenseKeyReceived.subscriberCount).toBe(1);
          expect((backend as any).backend).toBeInstanceOf(QueueBackend);

          adapter._onLicenseCallFailed.dispatch({});

          expect(clearTimeout).toHaveBeenCalledTimes(1);
          expect(clearTimeout).toHaveBeenLastCalledWith((backend as any).licenseLazyLoadingTimeoutHandle);
          expect(mockLicenseCall).toHaveBeenCalledTimes(0);
          expect(adapter.onLicenseCallFailed.subscriberCount).toBe(0);
          expect(adapter.onLicenseKeyReceived.subscriberCount).toBe(0);
          expect((backend as any).backend).toBeInstanceOf(NoOpBackend);
        });
      });

      describe('License key retrieved from adapter', () => {
        describe('empty key provided', () => {
          test('should resolve performLicenseCheck with denied and set a noop backend', () => {
            const adapter = new MockDeferredLicenseLoadingAdapterWithSupport(true);
            const mockLicenseCall = jest.fn().mockResolvedValue(granted);
            const backend = new LicenseCheckingBackend(
              licensingRequestNoKey,
              mockLicenseCall,
              ANALYTICS_BACKEND_BASE_URL,
              adapter
            );

            expect(mockLicenseCall).toHaveBeenCalledTimes(0);
            expect((backend as any).backend).toBeInstanceOf(QueueBackend);

            adapter._onLicenseKeyReceived.dispatch({licenseKey: ''});

            expect(clearTimeout).toHaveBeenLastCalledWith((backend as any).licenseLazyLoadingTimeoutHandle);
            expect(mockLicenseCall).toHaveBeenCalledTimes(0);
            expect(adapter.onLicenseCallFailed.subscriberCount).toBe(0);
            expect(adapter.onLicenseKeyReceived.subscriberCount).toBe(0);
            expect((backend as any).backend).toBeInstanceOf(NoOpBackend);
            const promise = backend.promise;
            expect(promise).toBeDefined();
            if (promise) {
              return promise.then((result) => {
                expect((backend as any).backend).toBeInstanceOf(NoOpBackend);
                expect(result.status).toEqual(LicensingResult.Denied);
              });
            }
          });
        });
        describe('invalid key provided', () => {
          test('should resolve performLicenseCheck with denied and set a noop backend', () => {
            const adapter = new MockDeferredLicenseLoadingAdapterWithSupport(true);
            const mockLicenseCall = jest.fn().mockResolvedValue(denied);
            const backend = new LicenseCheckingBackend(
              licensingRequestNoKey,
              mockLicenseCall,
              ANALYTICS_BACKEND_BASE_URL,
              adapter
            );

            expect(mockLicenseCall).toHaveBeenCalledTimes(0);
            expect((backend as any).backend).toBeInstanceOf(QueueBackend);

            adapter._onLicenseKeyReceived.dispatch({licenseKey: 'asdf'});

            expect(clearTimeout).toHaveBeenLastCalledWith((backend as any).licenseLazyLoadingTimeoutHandle);
            expect(mockLicenseCall).toHaveBeenCalledTimes(1);
            expect(adapter.onLicenseCallFailed.subscriberCount).toBe(0);
            expect(adapter.onLicenseKeyReceived.subscriberCount).toBe(0);
            const promise = backend.promise;
            expect(promise).toBeDefined();
            if (promise) {
              return promise.then((result) => {
                expect((backend as any).backend).toBeInstanceOf(NoOpBackend);
                expect(result.status).toEqual(LicensingResult.Denied);
              });
            }
          });
        });
        describe('valid key provided', () => {
          test('should resolve performLicenseCheck with granted and set a RemoteBackend', () => {
            const licenseKey = 'asdf';
            const adapter = new MockDeferredLicenseLoadingAdapterWithSupport(true);
            const mockLicenseCall = jest.fn().mockResolvedValue(granted);

            const backend = new LicenseCheckingBackend(
              licensingRequestNoKey,
              mockLicenseCall,
              ANALYTICS_BACKEND_BASE_URL,
              adapter
            );

            expect(mockLicenseCall).toHaveBeenCalledTimes(0);
            expect((backend as any).backend).toBeInstanceOf(QueueBackend);
            const flushToSpy = jest.spyOn((backend as any).backend, 'flushTo');

            backend.sendRequest({platform: 'ios', playerStartupTime: 100, pageLoadType: 1, pageLoadTime: 10});

            adapter._onLicenseKeyReceived.dispatch({licenseKey});

            expect(clearTimeout).toHaveBeenLastCalledWith((backend as any).licenseLazyLoadingTimeoutHandle);
            expect(mockLicenseCall).toHaveBeenCalledTimes(1);
            expect(adapter.onLicenseCallFailed.subscriberCount).toBe(0);
            expect(adapter.onLicenseKeyReceived.subscriberCount).toBe(0);
            const promise = backend.promise;
            expect(promise).toBeDefined();
            if (promise) {
              return promise.then((result) => {
                expect((backend as any).backend).toBeInstanceOf(RemoteBackend);
                expect(result.status).toEqual(LicensingResult.Granted);
                expect(flushToSpy).toHaveBeenCalledTimes(1);
                expect(flushToSpy).toHaveBeenCalledWith(expect.any(RemoteBackend));
              });
            }
          });
          test('should flush values with the new license key', () => {
            const licenseKey = 'asdf';
            const adapter = new MockDeferredLicenseLoadingAdapterWithSupport(true);
            const mockLicenseCall = jest.fn().mockResolvedValue(granted);
            const sendRequestMock = jest.fn();

            (AnalyticsCall as any).mockImplementation(() => {
              return {
                sendRequest: sendRequestMock,
              };
            });

            const backend = new LicenseCheckingBackend(
              licensingRequestNoKey,
              mockLicenseCall,
              ANALYTICS_BACKEND_BASE_URL,
              adapter
            );

            const sample = {platform: 'ios', playerStartupTime: 100, pageLoadType: 1, pageLoadTime: 10};
            backend.sendRequest(sample);
            adapter._onLicenseKeyReceived.dispatch({licenseKey});

            expect(mockLicenseCall).toHaveBeenCalledTimes(1);

            if (backend.promise) {
              return backend.promise.then((result) => {
                expect(sendRequestMock).toHaveBeenCalledWith({...sample, key: 'asdf'}, noOp);
              });
            }
          });
        });
      });
    });
  });
});
