import {BackendFactory} from '../js/core/BackendFactory';
import {DebuggingBackend} from '../js/core/DebuggingBackend';
import {LongPlaytimeFilteringBackend} from '../js/core/LongPlaytimeFilteringBackend';
import {NoOpBackend} from '../js/core/NoOpBackend';
import {RemoteBackend} from '../js/core/RemoteBackend';
import {AnalyticsConfig} from '../js/types/AnalyticsConfig';
import {LicensingResponse, LicensingResult} from '../js/types/LicensingRequest';
import {LicenseCall} from '../js/utils/LicenseCall';
import {ANALYTICS_BACKEND_BASE_URL} from '../js/utils/Settings';
import {MockDeferredLicenseLoadingAdapterWithSupport} from './utils/MockDeferredLicenseLoadingAdapterWithSupport';

jest.mock('../js/core/RemoteBackend');

jest.mock('../js/utils/LicenseCall');

const granted: LicensingResponse = {
  status: LicensingResult.Granted,
  message: 'here you go',
};

const validLicensingRequest = {
  domain: 'foo.com',
  key: '81354CDA-0EE6-4EDB-9A12-9EC7F80BBFE8',
  version: `1.2.3`,
};

const adapter = new MockDeferredLicenseLoadingAdapterWithSupport(true);

beforeEach(() => {
  (RemoteBackend as any).mockClear();
});

describe('BackendFactory', () => {
  beforeAll(() => {
    (LicenseCall as any).mockResolvedValue(granted);
  });

  test('it creates a RemoteBackend and makes a LicenseCall', async () => {
    const config: AnalyticsConfig = {
      key: 'someKey',
    };

    await new BackendFactory().createBackend(config, validLicensingRequest, adapter);

    expect(RemoteBackend).toHaveBeenCalledWith(
      true,
      ANALYTICS_BACKEND_BASE_URL,
      '81354CDA-0EE6-4EDB-9A12-9EC7F80BBFE8'
    );
    expect(LicenseCall).toHaveBeenCalled();
  });

  test('it creates RemoteBackend with backendUrl from analyticsConfig', async () => {
    const config: AnalyticsConfig = {
      key: 'someKey',
      config: {
        backendUrl: 'http://my-backend.com',
      },
    };

    await new BackendFactory().createBackend(config, validLicensingRequest, adapter);

    expect(RemoteBackend).toHaveBeenCalledWith(true, 'http://my-backend.com', '81354CDA-0EE6-4EDB-9A12-9EC7F80BBFE8');
  });

  test('it is of type DebuggingBackend when setting debug = true in the config', () => {
    const config: AnalyticsConfig = {
      key: 'someKey',
      debug: true,
    };

    const backend = new BackendFactory().createBackend(config, validLicensingRequest, adapter);

    expect(backend).toBeInstanceOf(DebuggingBackend);
  });
  test('it is not of type DebuggingBackend when setting debug = true in the config', () => {
    const config: AnalyticsConfig = {
      key: 'someKey',
      debug: false,
    };

    const backend = new BackendFactory().createBackend(config, validLicensingRequest, adapter);

    expect(backend).not.toBeInstanceOf(DebuggingBackend);
  });

  test('it is of type LongPlaytimeFilteringBackend', () => {
    const config: AnalyticsConfig = {
      key: 'someKey',
    };

    const backend = new BackendFactory().createBackend(config, validLicensingRequest, adapter);
    expect(backend).toBeInstanceOf(LongPlaytimeFilteringBackend);
  });

  test('it is of type NoOpBackend when setting enabled = false in the config', () => {
    const config: AnalyticsConfig = {
      key: 'someKey',
      config: {enabled: false},
    };

    const backend = new BackendFactory().createBackend(config, validLicensingRequest, adapter) as any;

    expect(backend).toBeInstanceOf(NoOpBackend);
  });

  test('it is not NoOpBackend when setting enabled = true in the config', () => {
    const config: AnalyticsConfig = {
      key: 'someKey',
      config: {enabled: true},
    };
    const backend = new BackendFactory().createBackend(config, validLicensingRequest, adapter) as any;
    expect(backend).not.toBeInstanceOf(NoOpBackend);
  });

  test('it is not NoOpBackend when omitting enabled in the config (is enabled by default)', () => {
    const config: AnalyticsConfig = {
      key: 'someKey',
      config: {},
    };
    const backend = new BackendFactory().createBackend(config, validLicensingRequest, adapter) as any;
    expect(backend).not.toBeInstanceOf(NoOpBackend);
  });
  test('it is not NoOpBackend when omitting config (is enabled by default)', () => {
    const config: AnalyticsConfig = {
      key: 'someKey',
    };
    const backend = new BackendFactory().createBackend(config, validLicensingRequest, adapter) as any;
    expect(backend).not.toBeInstanceOf(NoOpBackend);
  });
});
