import {Analytics} from '../../js/core/Analytics';
import {AnalyticsConfig} from '../../js/types/AnalyticsConfig';
import {VERSION} from '../../js/Version';
import {MockInternalAdapterApi} from '../utils/MockInternalAdapterApi';
import {TestBackend} from '../utils/TestBackend';
import {TestBackendFactory} from '../utils/TestBackendFactory';

jest.mock('../../js/core/BackendFactory');

describe('Analytics', () => {
  const testBackend = new TestBackend();
  const backendFactory = new TestBackendFactory(testBackend);
  test('has a version property', () => {
    expect(Analytics.version).toBeDefined();
    expect(Analytics.version).toEqual('version');
  });

  test('default values for config get set in constructor', () => {
    const adapter = new MockInternalAdapterApi();
    const config: AnalyticsConfig = {
      key: 'asdf',
    };
    const analytics = new Analytics(config, adapter, backendFactory);

    expect((analytics as any).config).toEqual({
      key: 'asdf',
      isLive: false,
    });
  });

  test('takes origin from collector configuration and sends licensing request', () => {
    const adapter = new MockInternalAdapterApi();
    const config: AnalyticsConfig = {
      key: 'asdf',
      config: {
        origin: 'com.package.name',
      },
    };

    const testBackend = new TestBackend();
    const backendFactory = new TestBackendFactory(testBackend);
    const createBackendMock = jest.fn(backendFactory.createBackend);
    backendFactory.createBackend = createBackendMock;

    const analytics = new Analytics(config, adapter, backendFactory);

    expect(createBackendMock).toHaveBeenCalledWith(
      config,
      {key: 'asdf', domain: 'com.package.name', version: VERSION},
      adapter
    );
  });

  test('takes customUserId or UserId for the sample field customUserId', () => {
    const userId = 'userID';
    const userId2 = 'userID2';

    const adapter = new MockInternalAdapterApi();
    const config: AnalyticsConfig = {
      key: 'asdf',
      customUserId: userId,
    };

    const analytics = new Analytics(config, adapter, backendFactory);

    analytics.setConfigParameters();

    expect(analytics.sample.customUserId).toEqual(userId);

    delete config.customUserId;
    config.userId = userId2;
    analytics.setConfigParameters();
    expect(analytics.sample.customUserId).toEqual(userId2);
  });

  test('takes userId in preference before customUserId to set sample customUserID', () => {
    const userId = 'userID';
    const userId2 = 'userID2';

    const adapter = new MockInternalAdapterApi();
    const config: AnalyticsConfig = {
      key: 'asdf',
      customUserId: userId,
      userId: userId2,
    };

    const analytics = new Analytics(config, adapter, backendFactory);

    analytics.setConfigParameters();

    expect(analytics.sample.customUserId).toEqual(userId2);
  });

  describe('Merging Analytics Configuration', () => {
    test('config remains the same after sourceChange without a config', () => {
      const adapter = new MockInternalAdapterApi();
      adapter.sourceChange = jest.fn();
      const config: AnalyticsConfig = {
        key: 'analyticsLicenseKey',
        playerKey: 'playerLicenseKey',
        cdnProvider: 'akamai',
        videoId: 'videoID',
        title: 'title',
        userId: 'userID',
        customData1: 42,
        customData2: 'cd2',
        customData3: 'cd3',
        customData4: false,
        customData5: 'cd5',
        customData6: 'cd6',
        customData7: 'cd7',
        experimentName: 'myExperiment',
        isLive: true,
        config: {
          origin: 'com.package.name',
          enabled: true,
          backendUrl: 'test',
          cookiesEnabled: true,
        },
      };

      const analytics = new Analytics(config, adapter, backendFactory);

      // @ts-ignore
      analytics.sourceChange(undefined);

      const expectedConfig = {
        key: 'analyticsLicenseKey',
        playerKey: 'playerLicenseKey',
        cdnProvider: 'akamai',
        videoId: 'videoID',
        title: 'title',
        userId: 'userID',
        customData1: 42,
        customData2: 'cd2',
        customData3: 'cd3',
        customData4: false,
        customData5: 'cd5',
        customData6: 'cd6',
        customData7: 'cd7',
        experimentName: 'myExperiment',
        isLive: true,
        config: {
          origin: 'com.package.name',
          enabled: true,
          backendUrl: 'test',
          cookiesEnabled: true,
        },
      };

      expect(adapter.sourceChange).toHaveBeenCalledWith(expectedConfig, expect.any(Number));
    });

    test('config is partially merged', () => {
      const adapter = new MockInternalAdapterApi();
      adapter.sourceChange = jest.fn();
      const config: AnalyticsConfig = {
        key: 'analyticsLicenseKey',
        playerKey: 'playerLicenseKey',
        videoId: 'videoID',
        title: 'title',
        userId: 'userID',
        customData1: 42,
        customData2: 'cd2',
        customData3: 'cd3',
        customData6: 'cd6',
        isLive: true,
      };

      const analytics = new Analytics(config, adapter, backendFactory);

      analytics.sourceChange({
        cdnProvider: 'akamai',
        customData4: false,
        customData5: 'cd5',
        experimentName: 'experimentName',
      });

      const expectedConfig = {
        key: 'analyticsLicenseKey',
        playerKey: 'playerLicenseKey',
        cdnProvider: 'akamai',
        videoId: 'videoID',
        title: 'title',
        userId: 'userID',
        customData1: 42,
        customData2: 'cd2',
        customData3: 'cd3',
        customData4: false,
        customData5: 'cd5',
        customData6: 'cd6',
        experimentName: 'experimentName',
        isLive: true,
        config: {
          enabled: true,
        },
      };

      expect(adapter.sourceChange).toHaveBeenCalledWith(expectedConfig, expect.any(Number));
    });

    test('empty config is populated with data from sourceChange and gets passed to adapter', () => {
      const player = {};
      const adapter = new MockInternalAdapterApi();
      adapter.sourceChange = jest.fn();
      const config: AnalyticsConfig = {
        key: 'analyticsLicenseKey',
        playerKey: 'playerLicenseKey',
      };

      const analytics = new Analytics(config, adapter, backendFactory);

      analytics.sourceChange({
        cdnProvider: 'akamai',
        videoId: 'videoID',
        title: 'title',
        userId: 'userID',
        customData1: 42,
        customData2: 'cd2',
        customData3: 'cd3',
        customData4: false,
        customData5: 'cd5',
        customData6: 'cd6',
        customData7: 'cd7',
        experimentName: 'experimentName',
        isLive: true,
      });

      const expectedConfig = {
        key: 'analyticsLicenseKey',
        playerKey: 'playerLicenseKey',
        cdnProvider: 'akamai',
        videoId: 'videoID',
        title: 'title',
        userId: 'userID',
        customData1: 42,
        customData2: 'cd2',
        customData3: 'cd3',
        customData4: false,
        customData5: 'cd5',
        customData6: 'cd6',
        customData7: 'cd7',
        experimentName: 'experimentName',
        isLive: true,
        config: {
          enabled: true,
        },
      };

      expect(adapter.sourceChange).toHaveBeenCalledWith(expectedConfig, expect.any(Number));
    });
  });

  describe('Merging Analytics collector configuration', () => {
    test('overrides all attributes with new config attributes', () => {
      const adapter = new MockInternalAdapterApi();
      const config: AnalyticsConfig = {
        key: 'asdf',
        config: {
          origin: 'com.package.name',
          enabled: true,
          cookiesEnabled: true,
          backendUrl: 'TestBackend',
        },
      };
      const analytics = new Analytics(config, adapter, backendFactory);

      const newConfig: AnalyticsConfig = {
        key: 'asdf',
        config: {
          origin: 'testOrigin',
          enabled: false,
          cookiesEnabled: false,
          backendUrl: 'ChangedBackend',
        },
      };
      const mergedCollectorConfig = analytics.getMergedCollectorConfig(newConfig, config);

      expect(mergedCollectorConfig).toEqual({
        origin: 'testOrigin',
        enabled: false,
        cookiesEnabled: false,
        backendUrl: 'ChangedBackend',
      });
    });

    test('config is partially merged', () => {
      const adapter = new MockInternalAdapterApi();
      const config: AnalyticsConfig = {
        key: 'asdf',
        config: {
          origin: 'com.package.name',
          enabled: true,
          cookiesEnabled: true,
          backendUrl: 'TestBackend',
        },
      };
      const analytics = new Analytics(config, adapter, backendFactory);

      const newConfig: AnalyticsConfig = {
        key: 'asdf',
        config: {
          origin: 'testOrigin',
        },
      };
      const mergedCollectorConfig = analytics.getMergedCollectorConfig(newConfig, config);

      expect(mergedCollectorConfig).toEqual({
        origin: 'testOrigin',
        enabled: true,
        cookiesEnabled: true,
        backendUrl: 'TestBackend',
      });
    });
  });

  test('check that enabled gets set to true in the sourceChange config if enabled is undefined in the new config ', () => {
    const adapter = new MockInternalAdapterApi();
    adapter.sourceChange = jest.fn();
    const config: AnalyticsConfig = {
      key: 'asdf',
      config: {
        origin: 'com.package.name',
        enabled: false,
      },
    };

    const analytics = new Analytics(config, adapter, backendFactory);

    // @ts-ignore
    analytics.sourceChange(undefined);

    expect(adapter.sourceChange).toHaveBeenCalledWith(
      {
        key: 'asdf',
        isLive: false,
        config: {
          origin: 'com.package.name',
          enabled: false,
        },
      },
      expect.any(Number)
    );

    analytics.sourceChange({
      key: 'asdfs',
      config: {
        origin: 'TestOrigin',
        enabled: false,
      },
    });

    expect(adapter.sourceChange).toHaveBeenCalledWith(
      {
        key: 'asdfs',
        isLive: false,
        config: {
          origin: 'TestOrigin',
          enabled: false,
        },
      },
      expect.any(Number)
    );

    analytics.sourceChange({
      key: 'asdfs',
      config: {
        origin: 'TestOrigin',
      },
    });

    const expectedConfig = {
      key: 'asdfs',
      isLive: false,
      config: {
        origin: 'TestOrigin',
        enabled: true,
      },
    };

    expect(adapter.sourceChange).toHaveBeenCalledWith(expectedConfig, expect.any(Number));
  });

  describe('getIsLiveFromConfigOrPlaybackInfo', () => {
    test('returns the value from config if playbackInfo is not defined', () => {
      const adapter = new MockInternalAdapterApi();
      const analytics = new Analytics({}, adapter, backendFactory);

      expect(analytics.getIsLiveFromConfigOrPlaybackInfo({isLive: true}, undefined)).toBeTruthy();
      expect(analytics.getIsLiveFromConfigOrPlaybackInfo({isLive: false}, undefined)).toBeFalsy();
    });

    test('returns the value from the playbackInfo instead of the value from the config', () => {
      const adapter = new MockInternalAdapterApi();
      const analytics = new Analytics({}, adapter, backendFactory);

      expect(analytics.getIsLiveFromConfigOrPlaybackInfo({isLive: false}, true)).toBeTruthy();
      expect(analytics.getIsLiveFromConfigOrPlaybackInfo({isLive: true}, false)).toBeFalsy();
    });

    test('returns false if neither is defined', () => {
      const adapter = new MockInternalAdapterApi();
      const analytics = new Analytics({}, adapter, backendFactory);

      expect(analytics.getIsLiveFromConfigOrPlaybackInfo({key: 'asdf'}, undefined)).toBeFalsy();
    });

    test('dropped frames in samples are differential', () => {
      const adapter = new MockInternalAdapterApi();
      const analytics = new Analytics({}, adapter, backendFactory);
      const getCurrentPlaybackInfoSpy = jest.spyOn(adapter, 'getCurrentPlaybackInfo');

      // initial value after setup should be 0
      analytics.setupSample();
      expect(analytics.sample.droppedFrames).toEqual(0);

      // 12 (new value) - 0 (previous value) => 12
      getCurrentPlaybackInfoSpy.mockReturnValue({droppedFrames: 12});
      analytics.setPlaybackInfoFromAdapter();
      expect(analytics.sample.droppedFrames).toEqual(12);

      // 12 (new value) - 12 (previous value) => 0 (adapter remembers previous value)
      analytics.setPlaybackInfoFromAdapter();
      expect(analytics.sample.droppedFrames).toEqual(0);

      // 16 (new value) - 12 (previous value) => 4
      getCurrentPlaybackInfoSpy.mockReturnValue({droppedFrames: 16});
      analytics.setPlaybackInfoFromAdapter();
      expect(analytics.sample.droppedFrames).toEqual(4);

      // case when e.g. source is change
      // 0 (new value) - 18 (previous value) => 0 (value should never be negative)
      getCurrentPlaybackInfoSpy.mockReturnValue({droppedFrames: 0});
      analytics.setPlaybackInfoFromAdapter();
      expect(analytics.sample.droppedFrames).toEqual(0);
    });
  });
});
