import {DownloadSpeedMeter} from '../../js/core/DownloadSpeedMeter';
import {SpeedMeasurement} from '../../js/types/SpeedMeasurement';
import {Converter} from '../../js/utils/Converter';

class MeterTest extends DownloadSpeedMeter {
  getMeasures(): SpeedMeasurement[] {
    return this.measures;
  }
}

describe('DownloadSpeedMeter', () => {
  describe('addMeasurement', () => {
    test('it adds the measurement to the measures list', () => {
      const meter = new MeterTest();
      meter.addMeasurement({
        duration: 1,
        size: 100,
        timeToFirstByte: 10,
        timestamp: new Date(),
        httpStatus: 200,
      });

      expect(meter.getMeasures().length).toEqual(1);
    });
    test('it ignores downloads that are faster than 300mbit (that probably came from cache)', () => {
      const meter = new MeterTest();
      meter.addMeasurement({
        duration: 1,
        size: 37500000, // 37.5 megabyte per second
        timeToFirstByte: 10,
        timestamp: new Date(),
        httpStatus: 200,
      });
      expect(meter.getInfo().segmentsDownloadCount).toEqual(0);
      expect(meter.getInfo().segmentsDownloadSize).toEqual(0);
      expect(meter.getInfo().segmentsDownloadTime).toEqual(0);
      expect(meter.getInfo().avgDownloadSpeed).toEqual(0);
      expect(meter.getInfo().minDownloadSpeed).toEqual(0);
      expect(meter.getInfo().maxDownloadSpeed).toEqual(0);
      expect(meter.getInfo().avgTimeToFirstByte).toEqual(0);
    });
    test('it does not ignore downloads that are a bit slower than 300mbit', () => {
      const meter = new MeterTest();
      meter.addMeasurement({
        duration: 1,
        size: 37400000, // 37.4 megabyte per second
        timeToFirstByte: 10,
        timestamp: new Date(),
        httpStatus: 200,
      });
      expect(meter.getInfo().segmentsDownloadCount).toEqual(1);
      expect(meter.getInfo().segmentsDownloadSize).toEqual(37400000);
      expect(meter.getInfo().segmentsDownloadTime).toEqual(1000);
      const bitsPerSecond = Converter.bytesToBits(37400000);
      expect(meter.getInfo().avgDownloadSpeed).toEqual(bitsPerSecond);
      expect(meter.getInfo().minDownloadSpeed).toEqual(bitsPerSecond);
      expect(meter.getInfo().maxDownloadSpeed).toEqual(bitsPerSecond);
      expect(meter.getInfo().avgTimeToFirstByte).toEqual(10000);
    });
    test('it adds cache hits regardless of http status code', () => {
      const meter = new MeterTest();
      meter.addMeasurement({
        duration: 1,
        size: 100,
        timeToFirstByte: 10,
        timestamp: new Date(),
        httpStatus: 304,
      });
      expect(meter.getMeasures().length).toEqual(1);
    });
    test('it ignores failed downloads', () => {
      const meter = new MeterTest();
      meter.addMeasurement({
        duration: 1,
        size: 100,
        timeToFirstByte: 10,
        timestamp: new Date(),
        httpStatus: 500,
      });
      expect(meter.getMeasures().length).toEqual(0);
    });
    test('it ignores failed downloads', () => {
      const meter = new MeterTest();
      meter.addMeasurement({
        duration: 1,
        size: 100,
        timeToFirstByte: 10,
        timestamp: new Date(),
        httpStatus: 404,
      });
      expect(meter.getMeasures().length).toEqual(0);
    });
  });
  describe('reset', () => {
    test('it clears the measures', () => {
      const meter = new MeterTest();
      meter.addMeasurement({
        duration: 1,
        size: 100,
        timeToFirstByte: 10,
        timestamp: new Date(),
        httpStatus: 200,
      });

      expect(meter.getMeasures().length).toEqual(1);
      meter.reset();
      expect(meter.getMeasures().length).toEqual(0);
    });
  });
  describe('getinfo', () => {
    describe('with empty meter', () => {
      const meter = new MeterTest();
      const info = meter.getInfo();
      expect(info.avgDownloadSpeed).toEqual(0);
      expect(info.minDownloadSpeed).toEqual(0);
      expect(info.maxDownloadSpeed).toEqual(0);
      expect(info.segmentsDownloadTime).toEqual(0);
      expect(info.segmentsDownloadSize).toEqual(0);
      expect(info.segmentsDownloadCount).toEqual(0);
      expect(info.avgTimeToFirstByte).toEqual(0);
    });
    const meter = new MeterTest();
    meter.addMeasurement({
      duration: 1,
      size: 1024, // one megabyte
      timeToFirstByte: 10,
      timestamp: new Date(),
      // Avg speed is 100/1 = 100b/s
      httpStatus: 200,
    });
    meter.addMeasurement({
      duration: 2,
      size: 1024, // one megabyte
      timeToFirstByte: 10,
      timestamp: new Date(),
      // Avg speed is 1024*8/2 = half a megabit
      httpStatus: 200,
    });
    const info = meter.getInfo();
    it('returns number of added segments as segmentsDownloaded', () => {
      expect(info.segmentsDownloadCount).toEqual(2);
    });

    it('returns sum of size as segmentsDownloadSize', () => {
      expect(info.segmentsDownloadSize).toEqual(2048);
    });
    it('returns sum of time as segmentsDownloadTime in milliseconds', () => {
      expect(info.segmentsDownloadTime).toEqual(3000);
    });
    it('returns average speed as avgDownloadSpeed in kbps', () => {
      expect(info.avgDownloadSpeed).toEqual(((1024 + 512) * 8) / 2);
    });
    it('returns the minimum speed as minDownloadSpeed in kbps', () => {
      expect(info.minDownloadSpeed).toEqual(512 * 8);
    });
    it('returns the maximum speed as maxDownloadSpeed in kbps', () => {
      expect(info.maxDownloadSpeed).toEqual(1024 * 8);
    });
    describe('segmentsDownloadTime', () => {
      it('rounds up to milliseconds', () => {
        const meter = new MeterTest();
        meter.addMeasurement({
          duration: 0.0003, // time in seconds
          size: 1024,
          timeToFirstByte: 10,
          timestamp: new Date(),
          httpStatus: 200,
        });

        expect(meter.getInfo().segmentsDownloadTime).toEqual(1);
      });
    });
    describe('avgTimeToFirstByte', () => {
      it('rounds up to milliseconds', () => {
        const meter = new MeterTest();
        meter.addMeasurement({
          duration: 0.0003, // time in seconds
          size: 1024,
          timeToFirstByte: 0.0003,
          timestamp: new Date(),
          httpStatus: 200,
        });

        expect(meter.getInfo().avgTimeToFirstByte).toEqual(1);
      });
    });
  });
});
