import {Backend} from '../../js/core/Backend';
import {LongPlaytimeFilteringBackend} from '../../js/core/LongPlaytimeFilteringBackend';
import {Sample} from '../../js/types/Sample';

describe('LongPlaytimeFilteringBackend', () => {
  const mockBackend: Backend = {
    sendRequest: jest.fn(),
    sendUnloadRequest: jest.fn(),
    sendRequestSynchronous: jest.fn(),
    sendAdRequest: jest.fn(),
  };
  const backend = new LongPlaytimeFilteringBackend(mockBackend);
  beforeEach(() => {
    jest.resetAllMocks();
  });

  interface TestCase {
    name: string;
    sample: any;
    shouldCallNext: boolean;
  }

  // Table driven tests that have to be run for each of the 4 methods of Backend.ts
  const cases: TestCase[] = [
    {name: 'it calls underlying backend with normal played', sample: {played: 300}, shouldCallNext: true},
    {name: 'it doesnt call underlying if sample.played is above 90000', sample: {played: 90001}, shouldCallNext: false},
    {name: 'it doesnt call underlying if sample.played is 90000', sample: {played: 90000}, shouldCallNext: false},
    {name: 'it calls underlying if sample.played is 89999', sample: {played: 89999}, shouldCallNext: true},
    {name: 'it calls underlying if sample.played is undefined', sample: {}, shouldCallNext: true},
    {name: 'it calls underlying if sample.played is 0', sample: {played: 0}, shouldCallNext: true},
    {name: 'it calls underlying if sample.played is negative', sample: {}, shouldCallNext: true},
  ];

  describe('sendRequest', () => {
    cases.forEach((t: TestCase) => {
      test(t.name, () => {
        backend.sendRequest(t.sample as Sample);
        if (t.shouldCallNext) {
          expect(mockBackend.sendRequest).toHaveBeenCalledWith(t.sample);
        } else {
          expect(mockBackend.sendRequest).not.toHaveBeenCalled();
        }
      });
    });
  });
  describe('sendUnloadRequest', () => {
    cases.forEach((t: TestCase) => {
      test(t.name, () => {
        backend.sendUnloadRequest(t.sample as Sample);
        if (t.shouldCallNext) {
          expect(mockBackend.sendUnloadRequest).toHaveBeenCalledWith(t.sample);
        } else {
          expect(mockBackend.sendUnloadRequest).not.toHaveBeenCalled();
        }
      });
    });
  });
  describe('sendRequestSynchronous', () => {
    cases.forEach((t: TestCase) => {
      test(t.name, () => {
        backend.sendRequestSynchronous(t.sample as Sample);
        if (t.shouldCallNext) {
          expect(mockBackend.sendRequestSynchronous).toHaveBeenCalledWith(t.sample);
        } else {
          expect(mockBackend.sendRequestSynchronous).not.toHaveBeenCalled();
        }
      });
    });
  });
  describe('sendAdRequest', () => {
    test('calls underlying backend', () => {
      const sample = {foo: 'bar'};
      backend.sendAdRequest(sample);
      expect(mockBackend.sendAdRequest).toHaveBeenCalledWith(sample);
    });
  });
});
