import {BITMOVIN_ANALYTICS_UUID_COOKIE_NAME, SessionPersistenceHandler} from '../../js/core/SessionPersistenceHandler';
import * as Utils from '../../js/utils/Utils';

describe('SessionPersistenceHandler', () => {
  beforeEach(() => {
    deleteAllCookies();
  });

  it('setup userId when no configuration exists', () => {
    const spyOnUtilsGetCookie = jest.spyOn(Utils, 'getCookie');
    const spyOnUtilsGenerateUUID = jest.spyOn(Utils, 'generateUUID');
    const spyOnDocCookieSetter = jest.spyOn(document, 'cookie', 'set');

    const sessionPersistenceHandler = new SessionPersistenceHandler({});

    expect(sessionPersistenceHandler.userId).not.toBeNull();
    expect(sessionPersistenceHandler.userId.length).toBeGreaterThan(0);
    expect(spyOnUtilsGetCookie).toBeCalledWith(BITMOVIN_ANALYTICS_UUID_COOKIE_NAME);
    expect(spyOnUtilsGetCookie).toReturnWith('');
    expect(spyOnUtilsGenerateUUID).toBeCalled();
    expect(spyOnDocCookieSetter).toBeCalledWith(
      `${BITMOVIN_ANALYTICS_UUID_COOKIE_NAME}=${sessionPersistenceHandler.userId}; path=/; expires="Mon, 31 Dec 2029 23:00:00 GMT"; max-age=315400000`
    );

    spyOnUtilsGetCookie.mockRestore();
    spyOnUtilsGenerateUUID.mockRestore();
    spyOnDocCookieSetter.mockRestore();
  });

  it('setup userId when configuration exists with cookies enabled', () => {
    const spyOnUtilsGetCookie = jest.spyOn(Utils, 'getCookie');
    const spyOnUtilsGenerateUUID = jest.spyOn(Utils, 'generateUUID');
    const spyOnDocCookieSetter = jest.spyOn(document, 'cookie', 'set');

    const sessionPersistenceHandler = new SessionPersistenceHandler({config: {cookiesEnabled: true}});

    expect(sessionPersistenceHandler.userId).not.toBeNull();
    expect(sessionPersistenceHandler.userId.length).toBeGreaterThan(0);
    expect(spyOnUtilsGetCookie).toBeCalledWith(BITMOVIN_ANALYTICS_UUID_COOKIE_NAME);
    expect(spyOnUtilsGetCookie).toReturnWith('');
    expect(spyOnUtilsGenerateUUID).toBeCalled();
    expect(spyOnDocCookieSetter).toBeCalledWith(
      `${BITMOVIN_ANALYTICS_UUID_COOKIE_NAME}=${sessionPersistenceHandler.userId}; path=/; expires="Mon, 31 Dec 2029 23:00:00 GMT"; max-age=315400000`
    );

    spyOnUtilsGetCookie.mockRestore();
    spyOnUtilsGenerateUUID.mockRestore();
    spyOnDocCookieSetter.mockRestore();
  });

  it('setup userId when configuration exists with cookies enabled and cookies domain', () => {
    const spyOnUtilsGetCookie = jest.spyOn(Utils, 'getCookie');
    const spyOnUtilsGenerateUUID = jest.spyOn(Utils, 'generateUUID');
    const spyOnDocCookieSetter = jest.spyOn(document, 'cookie', 'set');

    const sessionPersistenceHandler = new SessionPersistenceHandler({
      config: {cookiesEnabled: true, cookiesDomain: '.testDomain'},
    });

    expect(sessionPersistenceHandler.userId).not.toBeNull();
    expect(sessionPersistenceHandler.userId.length).toBeGreaterThan(0);
    expect(spyOnUtilsGetCookie).toBeCalledWith(BITMOVIN_ANALYTICS_UUID_COOKIE_NAME);
    expect(spyOnUtilsGetCookie).toReturnWith('');
    expect(spyOnUtilsGenerateUUID).toBeCalled();
    expect(spyOnDocCookieSetter).toBeCalledWith(
      `${BITMOVIN_ANALYTICS_UUID_COOKIE_NAME}=${sessionPersistenceHandler.userId}; path=/; expires="Mon, 31 Dec 2029 23:00:00 GMT"; max-age=315400000; domain=.testDomain`
    );

    spyOnUtilsGetCookie.mockRestore();
    spyOnUtilsGenerateUUID.mockRestore();
    spyOnDocCookieSetter.mockRestore();
  });

  it('setup userId when configuration exists with cookies disabled', () => {
    const spyOnUtilsGetCookie = jest.spyOn(Utils, 'getCookie');
    const spyOnUtilsGenerateUUID = jest.spyOn(Utils, 'generateUUID');
    const spyOnDocCookieSetter = jest.spyOn(document, 'cookie', 'set');

    const sessionPersistenceHandler = new SessionPersistenceHandler({config: {cookiesEnabled: false}});

    expect(sessionPersistenceHandler.userId).not.toBeNull();
    expect(sessionPersistenceHandler.userId.length).toBeGreaterThan(0);
    expect(spyOnUtilsGetCookie).toBeCalledWith(BITMOVIN_ANALYTICS_UUID_COOKIE_NAME);
    expect(spyOnUtilsGetCookie).toReturnWith('');
    expect(spyOnUtilsGenerateUUID).toBeCalled();
    expect(spyOnDocCookieSetter).not.toBeCalled();

    spyOnUtilsGetCookie.mockRestore();
    spyOnUtilsGenerateUUID.mockRestore();
    spyOnDocCookieSetter.mockRestore();
  });

  it('setup userId when configuration exists with cookies disabled and deletes existing one', () => {
    document.cookie = `${BITMOVIN_ANALYTICS_UUID_COOKIE_NAME}=existingUserId`;
    const spyOnUtilsGetCookie = jest.spyOn(Utils, 'getCookie');
    const spyOnUtilsGenerateUUID = jest.spyOn(Utils, 'generateUUID');
    const spyOnDocCookieSetter = jest.spyOn(document, 'cookie', 'set');

    const sessionPersistenceHandler = new SessionPersistenceHandler({config: {cookiesEnabled: false}});

    expect(sessionPersistenceHandler.userId).not.toBeNull();
    expect(sessionPersistenceHandler.userId).not.toBe('existingUserId');
    expect(sessionPersistenceHandler.userId.length).toBeGreaterThan(0);
    expect(spyOnUtilsGetCookie).toBeCalledWith(BITMOVIN_ANALYTICS_UUID_COOKIE_NAME);
    expect(spyOnUtilsGetCookie).toReturnWith('');
    expect(spyOnUtilsGenerateUUID).toBeCalled();
    expect(spyOnDocCookieSetter).toBeCalledWith(
      `${BITMOVIN_ANALYTICS_UUID_COOKIE_NAME}=; expires=Thu, 01 Jan 1970 00:00:00 UTC; path=/;`
    );

    spyOnUtilsGetCookie.mockRestore();
    spyOnUtilsGenerateUUID.mockRestore();
    spyOnDocCookieSetter.mockRestore();
  });

  it(`setup userId when cookie ${BITMOVIN_ANALYTICS_UUID_COOKIE_NAME} already exists`, () => {
    document.cookie = `${BITMOVIN_ANALYTICS_UUID_COOKIE_NAME}=existingUserId`;

    const spyOnUtilsGetCookie = jest.spyOn(Utils, 'getCookie');
    const spyOnUtilsGenerateUUID = jest.spyOn(Utils, 'generateUUID');
    const spyOnDocCookieSetter = jest.spyOn(document, 'cookie', 'set');

    const sessionPersistenceHandler = new SessionPersistenceHandler({});

    expect(sessionPersistenceHandler.userId).toBe('existingUserId');
    expect(spyOnUtilsGetCookie).toBeCalledWith(BITMOVIN_ANALYTICS_UUID_COOKIE_NAME);
    expect(spyOnUtilsGetCookie).toReturnWith('existingUserId');
    expect(spyOnUtilsGenerateUUID).not.toBeCalled();
    expect(spyOnDocCookieSetter).not.toBeCalled();

    spyOnUtilsGetCookie.mockRestore();
    spyOnUtilsGenerateUUID.mockRestore();
    spyOnDocCookieSetter.mockRestore();
  });
});

/**
 * @see https://stackoverflow.com/questions/179355/clearing-all-cookies-with-javascript
 */
function deleteAllCookies() {
  const cookies = document.cookie.split(';');

  for (let i = 0; i < cookies.length; i++) {
    const cookie = cookies[i];
    const eqPos = cookie.indexOf('=');
    const name = eqPos > -1 ? cookie.substr(0, eqPos) : cookie;
    document.cookie = name + '=;expires=Thu, 01 Jan 1970 00:00:00 GMT';
  }
}
