import {InternalAdapterAPI} from '../../js/adapters/internal/InternalAdapterAPI';
import {Analytics} from '../../js/core/Analytics';
import {NoOpBackend} from '../../js/core/NoOpBackend';
import {MockInternalAdapterApi} from '../utils/MockInternalAdapterApi';
import {TestBackend} from '../utils/TestBackend';
import {TestBackendFactory} from '../utils/TestBackendFactory';

describe('StateMachineCallback', () => {
  describe('video completion tracking', () => {
    test('playing event correctly tracks video completion', () => {
      const testBackend = new TestBackend();
      const adapter = new MockInternalAdapterApi();
      adapter.videoCompletionTracker.setVideoDuration(10);
      const backendFactory = new TestBackendFactory(testBackend);
      const analytics = new Analytics({key: 'license'}, adapter, backendFactory);

      setStartEndTimeInSample(adapter, 0, 2);
      adapter.stateMachineCallbacks.playing(2, 'playing');
      expect(testBackend.lastSendSample!.completedTotal).toEqual(0.2);
      expect(testBackend.lastSendSample!.completed).toEqual(0.2);
      setStartEndTimeInSample(adapter, 2, 9);
      adapter.stateMachineCallbacks.playing(9, 'playing');
      expect(testBackend.lastSendSample!.completedTotal).toEqual(0.9);
      expect(testBackend.lastSendSample!.completed).toEqual(0.7);
    });

    test('end resets video completion', () => {
      const testBackend = new TestBackend();
      const adapter = new MockInternalAdapterApi();
      adapter.videoCompletionTracker.setVideoDuration(10);
      const backendFactory = new TestBackendFactory(testBackend);
      const analytics = new Analytics({key: 'license'}, adapter, backendFactory);
      setStartEndTimeInSample(adapter, 0, 2);
      adapter.stateMachineCallbacks.playing(2, 'playing');
      expect(testBackend.lastSendSample!.completedTotal).toEqual(0.2);
      adapter.stateMachineCallbacks.end(4, 'paying');
      expect(adapter.videoCompletionTracker.getCompletionPercentage()).toEqual(0);
    });

    test('manual source change resets video completion', () => {
      const testBackend = new TestBackend();
      const adapter = new MockInternalAdapterApi();
      adapter.videoCompletionTracker.setVideoDuration(10);
      const backendFactory = new TestBackendFactory(testBackend);
      const analytics = new Analytics({key: 'license'}, adapter, backendFactory);
      setStartEndTimeInSample(adapter, 0, 2);
      adapter.stateMachineCallbacks.playing(2, 'playing');
      expect(testBackend.lastSendSample!.completedTotal).toEqual(0.2);
      adapter.stateMachineCallbacks.manualSourceChange({config: {}});
      expect(adapter.videoCompletionTracker.getCompletionPercentage()).toEqual(0);
    });
  });

  describe('quality change threshold', () => {
    test('after threshold exceeded send error sample', () => {
      const testBackend = new TestBackend();
      const adapter = new MockInternalAdapterApi();
      const backendFactory = new TestBackendFactory(testBackend);
      const analytics = new Analytics({key: 'license'}, adapter, backendFactory);

      adapter.stateMachineCallbacks.setVideoTimeStartFromEvent({currentTime: 1});
      adapter.stateMachineCallbacks.setVideoTimeEndFromEvent({currentTime: 3});
      adapter.stateMachineCallbacks.qualitychange(2, 'qualitychange');
      expect(testBackend.lastSendSample!.state).toEqual('qualitychange');

      adapter.stateMachineCallbacks.qualitychange(2, 'qualitychange');
      expect(testBackend.lastSendSample!.state).toEqual('qualitychange');

      adapter.stateMachineCallbacks.qualitychange(2, 'qualitychange');
      expect(testBackend.lastSendSample!.state).toEqual('error');
      expect(testBackend.lastSendSample!.videoTimeStart).toEqual(1000);
      expect(testBackend.lastSendSample!.videoTimeEnd).toEqual(3000);
    });

    test('manual source change resets qualityChangeCounter', () => {
      const testBackend = new TestBackend();
      const adapter = new MockInternalAdapterApi();
      const backendFactory = new TestBackendFactory(testBackend);
      const analytics = new Analytics({key: 'license'}, adapter, backendFactory);

      adapter.stateMachineCallbacks.qualitychange(2, 'qualitychange');
      expect(adapter.qualityChangeCounter).toBe(1);
      adapter.stateMachineCallbacks.qualitychange(3, 'qualitychange');
      expect(adapter.qualityChangeCounter).toBe(2);

      adapter.stateMachineCallbacks.manualSourceChange({config: {}});
      expect(adapter.qualityChangeCounter).toBe(0);
    });
  });

  describe('stop_collection', () => {
    it('disable Backend', () => {
      const testBackend = new TestBackend();
      const adapter = new MockInternalAdapterApi();
      const backendFactory = new TestBackendFactory(testBackend);
      const analytics = new Analytics({key: 'license'}, adapter, backendFactory);
      adapter.stateMachineCallbacks.stop_collecting();
      expect(analytics.backend).toBeInstanceOf(NoOpBackend);
    });
  });
});

// start and end in sec
function setStartEndTimeInSample(adapter: InternalAdapterAPI, start: number, end: number) {
  adapter.stateMachineCallbacks.setVideoTimeStartFromEvent({currentTime: start});
  adapter.stateMachineCallbacks.setVideoTimeEndFromEvent({currentTime: end});
}
