import {ErrorCode} from '../../js/enums/ErrorCode';
import {Event} from '../../js/enums/Event';
import {VideoStartFailedReason} from '../../js/enums/VideoStartFailedReason';
import ErrorEventObject from '../../js/types/ErrorEventObject';
import {defaultStateMachineCallbacks} from '../../js/types/StateMachineCallbacks';
import {MockedBitmovin8StateMachine} from '../utils/MockedBitmovin8StateMachine';
import {MockInternalAdapterApi} from '../utils/MockInternalAdapterApi';

describe('Video Startup Timeout', () => {
  it('ends in correct state and stops collecting', async () => {
    jest.setTimeout(1000 * 3);
    const stop_collectingSpy = jest.spyOn(defaultStateMachineCallbacks, 'stop_collecting');
    const stateMachine = new MockedBitmovin8StateMachine(
      {...defaultStateMachineCallbacks},
      {starttime: undefined},
      {videoStartupTimeoutMs: 300}
    );
    const adapter = new MockInternalAdapterApi(stateMachine);
    adapter.eventCallback(Event.SOURCE_LOADED, {
      currentTime: 0,
    });
    adapter.eventCallback(Event.PLAY, {
      currentTime: 0,
    });
    await new Promise((resolve) => {
      setTimeout(() => {
        resolve();
      }, 400);
    });

    expect(stateMachine.getCurrentState()).toBe('EXIT_BEFORE_VIDEOSTART');
    expect(stop_collectingSpy).toBeCalled();
  });

  it('contains timeout error data', async () => {
    jest.setTimeout(1000 * 3);
    const videoStartFailedSpy = jest.spyOn(defaultStateMachineCallbacks, 'videoStartFailed');
    const stateMachine = new MockedBitmovin8StateMachine(
      defaultStateMachineCallbacks,
      {starttime: undefined},
      {videoStartupTimeoutMs: 300}
    );
    const adapter = new MockInternalAdapterApi(stateMachine);
    adapter.eventCallback(Event.SOURCE_LOADED, {
      currentTime: 0,
    });
    adapter.eventCallback(Event.PLAY, {
      currentTime: 0,
    });
    await new Promise((resolve) => {
      setTimeout(() => {
        resolve();
      }, 400);
    });

    expect(stateMachine.getCurrentState()).toBe('EXIT_BEFORE_VIDEOSTART');
    expect(videoStartFailedSpy).toBeCalledWith(
      expect.objectContaining({
        reason: VideoStartFailedReason.Timeout,
        errorData: new ErrorEventObject(ErrorCode.VideoStartupTimeoutReached),
      }),
      true
    );
  });

  it('contains player error data', async () => {
    const errorSpy = jest.spyOn(defaultStateMachineCallbacks, 'error');
    const stateMachine = new MockedBitmovin8StateMachine(defaultStateMachineCallbacks, {starttime: undefined});
    const adapter = new MockInternalAdapterApi(stateMachine);
    adapter.eventCallback(Event.SOURCE_LOADED, {
      currentTime: 0,
    });
    adapter.eventCallback(Event.PLAY, {
      currentTime: 0,
    });
    const eventObject = {
      code: 1234,
      message: 'Player Error',
      currentTime: 0,
    };
    adapter.eventCallback(Event.ERROR, eventObject);

    expect(stateMachine.getCurrentState()).toBe('EXIT_BEFORE_VIDEOSTART');
    expect(errorSpy).toBeCalledWith(eventObject);
  });
});
